from setuptools import setup, find_packages
import os

setup(
    name='barcode_reader',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    version='1.3.7',
    packages=find_packages(),
    package_data={'barcode_reader.models': ['adder.pytorch.model']},
    include_package_data=True,
    install_requires=['torch>=1.13.1',
                      'torchvision',
                      'Pillow',
                      'zxing-cpp',
                      'piqa'
                    #   'autocapture_model'
                      ]
)
# cp /home/nathan/barcode-reader/dist/barcode_reader-1.0.25.tar.gz /var/lib/packages/barcode_reader-1.0.25.tar.gz
# python setup.py sdist
# pip install --extra-index-url https://pip.ml.prod.trulioo.com/nathan/dev --trusted-host https://pip.ml.prod.trulioo.com/ barcode_reader